/*
 * Ansteuerung einer 8x8 LED Matrix, gem. Kathode, Typ TOP-CA-788BS
 *
 * 21.06.2015	Florian Schffer, http://www.blafusel.de
 *
 * In einer extra Datei stehen die Textzeichen, die ausgegeben werden sollen. FuseBit EESAVE setzen, damit EPROM Inhalt beim Schreiben 
 * eines neuen Programmfiles nicht gelscht wird.
 * Im EEPROM stehen alle Spaltenbytes. Auch Leerspalten am Anfang (7) damit der Text von rechts ins Display scrollt.
 * 0xFF kennzeichnet EOF
 * Wartet 5 Sekunden auf Daten am USART, 8N1, 2400 Baud. Dabei leuchtet LED unten rechts. Whrend der bertragung flackert LED unten links
 * Alle Bytes werden in EPROM gesichert. Mit 0xFF = EOF
 * Unterste Zeile wird nicht wirklich genutzt, da fr 5x7 Font nicht erforderlich
 *
 * Datenblatt 8x8: http://www.ledtoplite.com/uploadfile/2014/0807/20140807032939345.pdf
 *
 * Achtung: ATtiny2313 und ATtiny2313A sind nicht Code-Kompatibel! 
 *
 */
 
 /*
 
PD0..6 = Kathode = Spalte 1..8, Pin 
PB0..7 = Anode = Zeile 1..7, Pin  mit Vorwiderstand
PA0		= Kathode = Spalte 8, Pin  
 
 Font: https://github.com/adafruit/Adafruit-GFX-Library/blob/master/glcdfont.c
 http://tinkering-arduino.de/blog/2014/03/24/font-generator-fur-das-8x8-led-matrix-display/
 
 */

#include <avr/io.h>
#include <util/delay.h>       	// definiert _delay_ms() 
#include <avr/interrupt.h>    	// IRQ Behandlung 
#include <avr/eeprom.h>				// EEPROM Zugriffe

#define UART_BAUD_RATE 2400L
#define UART_BAUD_CALC(UART_BAUD_RATE,F_CPU) ((F_CPU)/((UART_BAUD_RATE)*16L)-1)

volatile int16_t pos = 0;
volatile uint8_t flag = 0;

int main (void);

/**
   @brief   IRQ Routine wird bei erreichen des Vergleichswertes aufgerufen und sorgt dafr,
				da die nchste Datenspalte benutzt wird
*/
ISR (TIMER0_COMPA_vect)
{
	pos++;
	TCNT0 = 0;
}

/**
   @brief   IRQ Routine wird bei eingehenden Daten am USART aufgerufen
				Schreibt dann Byte in EEPROM
*/
SIGNAL (USART_RX_vect) 		
{
	uint8_t udr_data;

	udr_data= UDR; 			// Byte auf jeden Fall abholen, sonst Endlosinterrupt
	eeprom_busy_wait();
	eeprom_write_byte((uint8_t *) pos, udr_data); 		// Spaltendaten schreiben
	pos++;															// Zhler fr Spalte
	
	if (udr_data == 0xFF)		// Wenn EOF gesendet wurde => Ende Eingabe
		flag = 1;
}

/**
   @brief   Hauptroutine
   @param   none
   @return  Ende-Status
*/
int main(void)
{
	uint8_t i = 0;
	uint8_t col = 0;
	uint8_t daten = 0;

	PORTD = 0xFF;           // PORTD komplett Ein: Kathode
	PORTB = 0;           	// PORTB komplett Aus: Anode 
	PORTA &= ~(1 << PA0);	// Kathode A0 Aus
	DDRD = 0xFF;           	// PORTD komplett als Ausgang
	DDRB = 0xFF;				// PORTB komplett als Ausgang
	DDRA |= (1 << DDA0);		// Ausgang

	UCSRB |= (1 << RXEN );													// UART RX (empfangen) einschalten
	UCSRB |= (1 << RXCIE);													// Interruptauslsung fr eingehende Daten aktivieren
	UCSRC |= (1<<UCSZ0)|(1<<UCSZ1);										// Asynchron, 8N1
	UBRRH=(uint16_t)(UART_BAUD_CALC(UART_BAUD_RATE,F_CPU)>>8);	// Baudrate whlen
	UBRRL=(uint16_t)UART_BAUD_CALC(UART_BAUD_RATE,F_CPU);
	sei();																		// IRQs ein

	PORTA &= ~(1 << PA0);	// rechte Spalte Low => leuchtet LED unten rechts
	PORTB |= (1 << PB7);		// Reihe 
	_delay_ms(3000);			// Warte auf Eingabe

	while (!flag && pos)		// Solange Daten eingegeben wurden (pos) und nicht EOF
		_delay_ms(500);
	
	cli();						// IRQs aus
	UCSRB = 0;  				// UART aus
	pos = 0;

	TCCR0B = (1 << CS02) | (1 << CS00); 	// 8 Bit Timer, Prescaler CLK/1024
	OCR0A = 100;									// Vergleichswert, kleiner = schneller
	TIMSK = (1 << OCIE0A);
	sei();                     				// IRQs ein

	while (1)  // endlos
	{
	   PORTD = 0xFF;		// Alle Kathode Aus (invertierte Logik)
		PORTA |= (1 << PA0);
		
		eeprom_busy_wait();
		daten = eeprom_read_byte((uint8_t *) (pos + col)); 		// Spaltendaten auslesen
		if (daten == 0xFF)		// EOF Marke
		{
			pos = 0;
		}
		else
		{	
			if (col == 7)
			{
				PORTD = 0xFF;				// alle anderen Spalten High
				PORTA = ~(1 << PA0);		// rechte Spalte Low 
			}
			else
			{
				PORTA |= (1 << PA0);		// rechte Spalte High
				PORTD = ~(1 << col);		// Spalte (invertierte Logik: 0 = Spalte Ein)
			}	
			
			for (i = 0; i < 8; i++)			// Zeilenweise durchgehen und LED kurz leuchten lassen => Strombelastung immer gleich
			{
				PORTB = daten & (1 << i);
				_delay_us(260);				// etwas Schaltzeit belassen (1000 ms/64 Pixel/60 1/2Hz = 260 s)
			}
		}

		col++;
		if (col == 8)		// Wenn 8 Spalten der Matrix durchlaufen sind
			col = 0;
	}

	return 1;      // nie
}

